#include <stdio.h>
#include <stdlib.h>
#include <termios.h>

/*------------------------------------------------------------------------*/

/* Konstanten fuer die Farben definieren. */
#define COLOR_BLACK		0
#define COLOR_RED		1
#define COLOR_GREEN		2
#define COLOR_YELLOW	3
#define COLOR_BLUE		4
#define COLOR_MAGENTA	5
#define COLOR_CYAN		6
#define COLOR_WHITE		7

/*------------------------------------------------------------------------*/

struct termios origstate;

/* Alten Terminalzustand speichern (von stdin = id 0). */
void setup() {
	struct termios changedstate;
    if (tcgetattr(0, &origstate) != 0)
    printf("tcgetattr failed");
    changedstate = origstate;

    /* Neuen Terminalzustand setzen (von stdin = id 0). */
    changedstate.c_lflag &= ~ICANON; /* disable canonical input */
    changedstate.c_lflag &= ~ECHO;   /* disable output */
    changedstate.c_cc[VMIN] = 1;     /* min = 1 */
    changedstate.c_cc[VTIME] = 1;    /* time = 1/10 sec */
    if (tcsetattr(0, TCSANOW, &changedstate) != 0)
    printf("tcsetattr failed");
}

/* Alten Terminalzustand wiederherstellen (von stdin = id 0). */
void restore() {
    if (tcsetattr(0, TCSANOW, &origstate) != 0)
    printf("tcsetattr failed");
}

/*------------------------------------------------------------------------*/

/* Bildschirm loeschen. */
void clear_screen() {
	printf("\033[2J");
}

/* Farbe fuer Hintergrund setzen. */
void set_bgcolor(int color) {
	printf("\033[%dm", color + 40);
}

/* Farbe fuer Vordergrund setzen. */
void set_fgcolor(int color) {
	printf("\033[%dm", color + 30);
}

/* Cursor positionieren. */
void set_cursor(int x, int y) {
	printf("\033[%d;%dH", y + 1, x + 1);
}

/* Farben und Attribute zuruecksetzen. */
void reset() {
	printf("\033[0m");
}

/*------------------------------------------------------------------------*/

/* Spielfeld einlesen. */
void read_level(char *d, char ***f, int ***t, int *w, int *h, int *x, int *y) {

	FILE *fp = NULL;
	char ch;
	int i, j;

    if (!(fp = fopen(d, "r"))) {
        fprintf(stderr, "Keine weiteren Levels mehr gefunden!\n");
        exit(1);
    }

	/* Groesse des Spielfeldes einlesen. */
    fscanf(fp, "%d %d", w, h);
	while ( (ch = fgetc(fp)) != '\n'); 

	/* Speicher fuer Spielfeld und Zielpositionen reservieren. */
	(*f) = calloc(sizeof(char *), *w);
	(*t) = calloc(sizeof(int *), *w);
	for (i = 0; i < *w; i++) {
		(*f)[i] = calloc(sizeof(char), *h);
		(*t)[i] = calloc(sizeof(int), *h);
	}

	/* Zeilenweise die Daten einlesen. */
    for (i = 0; i < *h; i++) {
        for (j = 0; j < *w; j++) {

            /* Alle Zeichen einer Zeile lesen. */
            (*f)[j][i] = fgetc(fp);

			/* Position des Weihnachtsmannes bestimmen. */
			if ((*f)[j][i] == '.') {
				*x = j;
				*y = i;
				(*f)[j][i] = ' ';
			}

			/* Zielpositionen bestimmen. */
			if ((*f)[j][i] != '_') {
				(*t)[j][i] = 0;
			}
			else {
				(*f)[j][i] = ' ';
				(*t)[j][i] = 1;
			}
        }
        /* Newline einer Zeile lesen. */
        ch = fgetc(fp);
    }
	fclose(fp);
}

/* Spielfeld ausgeben. */
void print_board(char **f, int w, int h) {
	int i, j;
	for (i = 0; i < h; i++) {
		for (j = 0;  j < w; j++) {
			printf("%c", f[j][i]);
		}
		printf("\n");
	}
}

/* Spielfeld farbig ausgeben. */
void print_boardcolored(char **f, int **t, int w, int h, int x, int y, int l) {
	int i, j;

	clear_screen();

	for (i = 0; i < h; i++) {
		set_cursor(0, i);
		for (j = 0;  j < w; j++) {
			if (t[j][i] == 1) {
				set_bgcolor(COLOR_YELLOW);
			}
			else if (f[j][i] == '#') {
				set_bgcolor(COLOR_BLUE);
			}
			printf(" ");
			reset();
		}
		printf("\n");
	}
	
	/* Weihnachtsmann und Geschenke erst am Schluss zeichnen, */
	/* da sie sonst evtl. von den Zielpos. ueberschrieben werden. */
	for (i = 0; i < h; i++) {
		for (j = 0;  j < w; j++) {
			if ( (j == x) && (i == y) ) {
				set_bgcolor(COLOR_BLACK);
				set_cursor(j, i);
				printf(" ");
			}
			else if (f[j][i] == '*') {
				set_bgcolor(COLOR_RED);
				set_cursor(j, i);
				printf(" ");
			}
			reset();
		}
		printf("\n");
	}
	printf("\nLevel %02d\n", l);
	printf("\nSteuerung: j l i m, (q)uit\n");
}

/* Speicher fuer Brett freigeben. */
void discard(char ***f, int ***t, int w, int h) {
	int i;
	for (i = 0; i < w; i++) {
		free((*f)[i]);
		free((*t)[i]);
		(*f)[i] = NULL;
		(*t)[i] = NULL;
	}
	free(*f);
	free(*t);
	(*f) = NULL;
	(*t) = NULL;
}

/*------------------------------------------------------------------------*/

/* Ist ein Zug bzgl. der akt. Pos. um dx bzw. dy moeglich? */
int move_possible(char **f, int x, int y, int dx, int dy, int w, int h) {

	int x1 = x + dx;
	int y1 = y + dy;
	/* Eine Position weiter als die Zielposition berechnen! */
	int x2 = x + 2 * dx;
	int y2 = y + 2 * dy;

	/* Groesserer Schritt als 1 bzgl. einer Koordinate? */
	if (!( (dx == 0 && dy*dy == 1) || (dx*dx == 1 && dy == 0) )) {
		return 0;
	}

	/* Zielposition nicht im Feld? */
	if (x1 < 0 || x1 >= w || y1 < 0 || y1 >= h) {
		return 0;
	}

	/* Zielposition leer? */
	if ( f[x1][y1] == ' ') {
		return 1;
	}

	/* Ist Zielposition + 1 Feld weiter innerhalb des Feldes? */
	if (x2 < 0 || x2 >= w || y2 < 0 || y2 >= h) {
		return 0;
	}

	/* Ist bei (x1, y1) eine Kiste und bei (x2, y2) leer, dann Ok! */
	if (f[x1][y1] == '*' && f[x2][y2] == ' ') {
		return 1;
	}

	return 0;

} 

/* Falls Zug moeglich dann ausfuehren! */
void move_player(char **f, int *x, int *y, int dx, int dy, int w, int h) {

	int x1 = *x + dx;
	int y1 = *y + dy;

	int x2 = *x + 2 * dx;
	int y2 = *y + 2 * dy;

	if ((move_possible(f, *x, *y, dx, dy, w, h)) == 1) {
		if (f[x1][y1] != ' ') {
			/* Kiste verschieben. */
			f[x2][y2] = f[x1][y1];
			f[x1][y1] = ' ';
		}
		*x = x1; *y = y1;
	}	
}

/*------------------------------------------------------------------------*/

int level_finished(char **f, int **t, int w, int h) {
	int i, j;
	for (i = 0; i < h; i++) {
		for (j = 0;  j < w; j++) {

			if (f[j][i] == '*') {
				if (! (t[j][i] == 1) ) {
					return 0;
				}
			}

		}
	}
	return 1;
}

/*------------------------------------------------------------------------*/

int main() {

	int width, height; 	/* Breite und Hoehe des Spielfeldes. */
	int x, y; 			/* Aktuelle Position des Weihnachtsmannes. */
	int level;			/* Levelnummer. */
	char **field;		/* Spielfeld. */
	int **target;		/* Zielfelder. */
	char *datei;		/* Dateiname. */
	char ch;			/* Welche Taste? */

	level = 1;
	datei = calloc(sizeof(char),12);
	sprintf(datei, "level%02d.dat", level);	

	read_level(datei, &field, &target, &width, &height, &x, &y);
	print_boardcolored(field, target, width, height, x, y, level);

    /* Alten Terminalzustand speichern, neuen setzen */
	setup();
	while ((ch = getchar()) != 'q') {
	    if (ch == 'i') {
			move_player(field, &x, &y, 0, -1, width, height);
		    print_boardcolored(field, target, width, height, x, y, level);
		}
	    else if (ch == 'm') {
			move_player(field, &x, &y, 0, 1, width, height);
		    print_boardcolored(field, target, width, height, x, y, level);
		}
	    else if (ch == 'j') {
			move_player(field, &x, &y, -1, 0, width, height);
		    print_boardcolored(field, target, width, height, x, y, level);
		}
	    else if (ch == 'l') {
			move_player(field, &x, &y, 1, 0, width, height);
		    print_boardcolored(field, target, width, height, x, y, level);
		}
		
		if (level_finished(field, target, width, height) == 1) {
			level++;
			discard(&field, &target, width, height);
			sleep(1);
			printf("\nJuppieh, Level geschafft! Nchster Level wird geladen!\n");
			fflush(stdout);
			sleep(2);
			sprintf(datei, "level%02d.dat", level);	
			read_level(datei, &field, &target, &width, &height, &x, &y);
			print_boardcolored(field, target, width, height, x, y, level);
		}
	}
    /* Alten Terminalzustand wieder herstellen. */
	restore();

	return 0;
}
